# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import base64


def b64enc(value):
    """
    Encode a string using base64.

    :param value: Text to be encoded
    :type value: ``str``
    :return: Encoded text
    :rtype: ``str``

    Use this filter for k8s secrets. Values for secrets need to be encoded
    before the ``Secret`` resource is deployed to k8s.
    """
    return base64.b64encode(value.encode('utf-8')).decode('ASCII')


def rtrim(value):
    """
    Strip trailing whitespace.

    :param value: Text to be processed
    :type value: ``str``
    :return: Text without trailing whitespace
    :rtype: ``str``
    """
    return value.rstrip()


def secretmap(value):
    """
    Label path with ``{{secretmap}}``.

    :param value: Path
    :type value: ``str``
    :return: Labeled path
    :rtype: ``str``
    """
    return os.path.join('{{secretmap}}', value)


filters = {
    'b64enc': b64enc,
    'rtrim': rtrim,
    'secretmap': secretmap,
}



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import subprocess
from contextlib import suppress


class AmbiguousConfiguration(Exception):
    """Give warning that something is wrong with configuration."""

    pass


class Singleton(type):
    """
    Define the Singleton.
    """

    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class DeleteFile:
    """Delete file from the file system, ideally in secure manner."""

    def __init__(self, path):
        """
        Make an instance of :class:`narrenschiff.common.DeleteFile` class.

        :param path: Absolute path to the file that needs to be deleted
        :type path: ``str``
        :return: Void
        :rtype: ``None``
        """
        self.path = path

    def delete(self):
        """Delete the file."""
        try:
            subprocess.run(['shred', self.path])
        except Exception:
            self._placebo_delete(passes=3)

        os.remove(self.path)

    def _placebo_delete(self, passes=1):
        """
        Overwrite file before deletaion. This is executed in case the OS
        does not have the ``shred`` command line utility. This may not work for
        modern systems that use journaling file systems, copy-on-write file
        systems, wear leveling, or similar.

        :param passes: Number of times file will be overwritten
        :type passes: ``int``
        :return: Void
        :rtype: ``None``
        """
        length = os.path.getsize(self.path)
        with open(self.path, 'wb') as f:
            for _ in range(passes):
                f.seek(0)
                f.write(os.urandom(length))
                os.fsync(f)


def get_chest_file_path(location):
    """
    Check if there are duplicate chest files, and return paths if there are
    not.

    :param location: Relative path to course project directory
    :type location: ``str``
    :return: Absolute path to chest file
    :rtype: ``str``
    :raises: :class:`narrenschiff.common.AmbiguousConfiguration`
    """
    paths = ['chest.yaml', 'chest.yml']
    cwd = os.getcwd()
    candidate = []
    for path in paths:
        chest_file = os.path.join(cwd, location, path)
        if os.path.exists(chest_file):
            candidate.append(chest_file)

    candidate_length = len(candidate)
    if candidate_length > 1:
        raise AmbiguousConfiguration('Only one chest file can exist')
    if candidate_length == 0:
        raise AmbiguousConfiguration('Chest file does not exist')
    return candidate[0]


def flatten(lst):
    """
    Flatten list.

    :param lst: A list to be flattened
    :type lst: ``list``
    :return: Flattened list
    :rtype: ``list``
    """
    flattened = []
    for element in lst:
        if isinstance(element, str):
            flattened.append(element)
        else:
            with suppress(TypeError):
                flattened.extend(element)
    return flattened


def is_yaml(filename):
    return bool(re.search(r'ya?ml$', filename, re.I))


def is_jinja(filename):
    return bool(re.search(r'j(inja)?2$', filename, re.I))



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import base64

import yaml
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import modes
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class AES256Cipher:
    """Encode and/or decode strings."""

    def __init__(self, keychain):
        """
        Construct :class:`narrenschiff.chest.AES256Cypher`.

        :param keychain: Object containing key and spice
        :type keychain: :class:`narrenschiff.chest.Keychain`
        :return: Void
        :rtype: ``None``
        """
        self.password = keychain.key.encode('utf-8')
        self.salt = keychain.spice.encode('utf-8')

        self.backend = default_backend()
        self.key = self.pbkdf2()

    def encrypt(self, plaintext):
        """
        Encrypt plaintext.

        :param plaintext: Plaintext
        :type plaintext: ``str``
        :return: Ciphertext
        :rtype: ``str``
        """
        iv = os.urandom(16)
        cipher = Cipher(
            algorithms.AES(self.key),
            modes.CBC(iv),
            backend=self.backend
        )

        encryptor = cipher.encryptor()
        padder = padding.PKCS7(256).padder()
        plaintext = plaintext.rstrip().encode('utf-8')
        padded = padder.update(plaintext) + padder.finalize()
        ciphertext = encryptor.update(padded) + encryptor.finalize()
        return base64.b64encode(iv + ciphertext)

    def decrypt(self, ciphertext):
        """
        Decrypt ciphertext.

        :param ciphertext: Ciphertext
        :type ciphertext: ``str``
        :return: Plaintext
        :rtype: ``str``
        """
        ciphertext = base64.b64decode(ciphertext.rstrip())
        iv, ciphertext = ciphertext[:16], ciphertext[16:]
        cipher = Cipher(
            algorithms.AES(self.key),
            modes.CBC(iv),
            backend=self.backend
        )

        decryptor = cipher.decryptor()
        unpadder = padding.PKCS7(256).unpadder()
        plaintext = decryptor.update(ciphertext) + decryptor.finalize()
        unpadded = unpadder.update(plaintext) + unpadder.finalize()
        return unpadded.decode('utf-8')

    def pbkdf2(self):
        """
        Derive a 32 bytes (256 bits) key.

        :return: Password hash
        :rtype: ``byte string``
        """
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=self.salt,
            iterations=100000,
            backend=self.backend
        )
        return kdf.derive(self.password)


class Chest:
    """Manipulate chest file."""

    def __init__(self, keychain, path):
        """
        Unlock treasure chest.

        :param keychain: Object containing key and spice
        :type keychain: :class:`narrenschiff.chest.Keychain`
        :param path: Path to the chest file
        :type path: ``str``
        :return: Void
        :rtype: ``None``
        """
        self.keychain = keychain
        self.path = path
        self.chest = self.load_chest_file()

    def load_chest_file(self):
        """
        Load chest file with encrypted values.

        :return: Serialized YAML object
        :rtype: ``dict``
        """
        with open(self.path, 'r') as f:
            chest = yaml.load(f, Loader=yaml.FullLoader)
        return chest if chest else {}

    def update(self, variable, value):
        """
        Add or update chest file.

        :param variable: Variable name to update
        :type variable: ``str``
        :param value: Value of the variable
        :type value: ``str``
        :return: Void
        :rtype: ``None``
        """
        cipher = AES256Cipher(self.keychain)
        self.chest[variable] = cipher.encrypt(value).decode('utf-8')

        with open(self.path, 'w') as f:
            f.write(yaml.dump(self.chest))

    def show(self, variable):
        """
        Show decrypted value of the variable.

        :param variable: Variable name
        :type variable: ``str``
        :param value: Value of the variable
        :type value: ``str``
        :return: Decrypted value
        :rtype: ``str``
        """
        cipher = AES256Cipher(self.keychain)
        return cipher.decrypt(self.chest[variable])



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import click

from narrenschiff.config import Keychain
from narrenschiff.templating import SecretmapVars
from narrenschiff.secretmap import Secretmap
from narrenschiff.secretmap import CourseLocationError


@click.group()
@click.pass_context
def secretmap(ctx):
    """
    Encrypt, decrypt, and edit files.
    \f

    :return: Void
    :rtype: ``None``
    """
    ctx.obj = Keychain()


@secretmap.command()
@click.option('--source', help='Source filepath for encryption.')
@click.option(
    '--destination',
    help=('Destination of the encrypted file.'
          ' Path relative to course project directory.')
)
@click.option('--treasure', help='Variable name.')
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def stash(keychain, source, destination, treasure, location):
    """
    Encrypt and stash file.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param source: Source filepath for encryption
    :type source: ``str``
    :param destination: Destination filepath of the encrypted file
    :type destination: ``str``
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param location: Location of the secretmap file
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    secretmap = Secretmap(keychain=keychain, directory=location)
    secretmap.upsert(src=source, dest=destination, treasure=treasure)


@secretmap.command()
@click.option('--destination', help='Destination of the encrypted file.')
@click.option('--treasure', help='Variable name.')
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def loot(keychain, destination, treasure, location):
    """
    Decrypt file from stash.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param destination: Destination filepath of the decrypted file
    :type destination: ``str``
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param location: Location of the secretmap file
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    secretmap = Secretmap(keychain=keychain, directory=location)
    secretmap.decrypt(dest=destination, treasure=treasure)


@secretmap.command()
@click.option('--treasure', help='Variable name from the secretmap file.')
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def peek(keychain, treasure, location):
    """
    Print content of the encrypted file to STDOUT.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param location: Location of the secretmap file
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    secretmap = Secretmap(keychain=keychain, directory=location)
    secretmap.peek(treasure)


@secretmap.command()
@click.option(
    '--treasure',
    help='Variable name from the secretmap file of the file you want to edit.'
)
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def alter(keychain, treasure, location):
    """
    Decrypt and edit the file. After edits, the file is encrypted again.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param location: Location of the secretmap file
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    secretmap = Secretmap(keychain=keychain, directory=location)
    secretmap.edit(treasure)


@secretmap.command()
@click.option('--treasure', help='Variable name from the secretmap file.')
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def destroy(keychain, treasure, location):
    """
    Delete secretmap file and corresponding key in the secretmap.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param location: Location of the secretmap file
    :type location: ``str``
    :return: Void
    :rtype: ``None``

    **You cannot undo this action!**
    """
    secretmap = Secretmap(keychain=keychain, directory=location)
    secretmap.destroy(treasure)


@secretmap.command()
@click.option('--location', help='Relative path to course project directory.')
@click.option('--match', help='Pattern you are searching for.')
@click.pass_obj
def search(keychain, location, match):
    """
    Search for a pattern in secretmaps.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param location: Location of the secretmap file
    :type location: ``str``
    :param match: Pattern you are looking for
    :type match: ``str``
    :return: Void
    :rtype: ``None``
    """
    if not os.path.isdir(location):
        raise CourseLocationError

    vars = SecretmapVars(location).load_vars()

    for treasure in vars[0].keys():
        secretmap = Secretmap(keychain=keychain, directory=location)
        secretmap.find(match, treasure)


@secretmap.command()
@click.option('--location', help='Relative path to course project directory.')
@click.argument('secretmaps', nargs=2)
@click.pass_obj
def diff(keychain, location, secretmaps):
    """
    Compare secretmaps line by line.

    It can only compare secretmaps in the same course project.
    \f

    :param location: Location of the secretmap file
    :type location: ``str``
    :param secretmaps: Two secretmaps that should be compared
    :type secretmaps: ``str``
    :return: Void
    :rtype: ``None``
    """
    secretmap = Secretmap(keychain=keychain, directory=location)
    secretmap.diff(secretmaps)



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import uuid
import shutil
import difflib
import tempfile
import subprocess
from contextlib import suppress

import yaml
import click

from narrenschiff.chest import AES256Cipher
from narrenschiff.common import Singleton
from narrenschiff.common import DeleteFile
from narrenschiff.log import NarrenschiffLogger


logger = NarrenschiffLogger()


class CourseLocationError(Exception):
    """Raise exception if course is not found."""

    pass


class Secretmap(metaclass=Singleton):
    """Manage secret maps. Secret maps are paths to encrypted files."""

    FILENAME = 'secretmap.yaml'

    def __init__(self, keychain, directory):
        """
        Construct a :class:`narrenschiff.secretmap.Secretmap`.

        :param keychain: Keychain contains key and spice
        :type keychain: :class:`narrenschiff.chest.Keychain`
        :param directory: Path to directory containing secretmap
        :type directory: ``str``
        :return: Void
        :rtype: ``None``
        """
        if os.path.isfile(directory):
            self.directory = os.path.dirname(directory)
        elif os.path.isdir(directory):
            self.directory = directory
        else:
            raise CourseLocationError

        self.filepath = os.path.join(directory, Secretmap.FILENAME)
        self.keychain = keychain
        self.tmp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))

    def upsert(self, src, dest, treasure):
        """
        Encrypts file and inserts data to config file.

        :param src: Source filepath for encryption
        :type src: ``str``
        :param dest: Destination filepath of the encrypted file
        :type dest: ``str``
        :param treasure: Name of the variable
        :type treasure: ``str``
        :return: Void
        :rtype: ``None``
        """
        with open(src, 'r') as f:
            cipher = AES256Cipher(self.keychain)
            enc_file_core = cipher.encrypt(f.read()).decode('utf-8')

        dest_abspath = os.path.abspath(os.path.join(self.directory, dest))
        try:
            os.makedirs(os.path.dirname(dest_abspath), 0o755)
        except FileExistsError:
            pass

        with open(dest_abspath, 'w') as f:
            f.write(enc_file_core)

        config = self._read_config()
        config[treasure] = dest
        self._write_config(config)

    def decrypt(self, dest, treasure):
        """
        Decrypts file and stores it to given destination.

        :param dest: Destination filepath of the decrypted file
        :type dest: ``str``
        :param treasure: Name of the variable
        :type treasure: ``str``
        :return: Void
        :rtype: ``None``
        """
        with open(dest, 'w') as f:
            f.write(self._decrypt(treasure))

    def peek(self, treasure):
        """
        Print encrypted file to STDOUT.

        :param treasure: Name of the secretmap variable
        :type treasure: ``str``
        :return: Void
        :rtype: ``None``
        """
        click.echo(self._decrypt(treasure))

    def find(self, match, treasure):
        """
        Match a pattern in a treasure and print to STDOUT.

        :param match: Pattern to match
        :type match: ``str``
        :param treasure: Name of the secretmap variable
        :type treasure: ``str``
        :return: Void
        :rtype: ``None``
        """
        secretmap = self._decrypt(treasure).split('\n')

        logger.debug(f'Searching for "{match}"')
        for index, line in enumerate(secretmap, start=1):
            candidate = re.search(match, line)
            if candidate:
                result = candidate.group()
                prefix = (f'\033[35m{treasure}\033[0m:\033[32m{index}\033[0m')
                formatted = line.replace(result, f'\033[31m{result}\033[0m')
                print(f'{prefix}:{formatted}')

    def diff(self, secretmaps):
        """
        Compare secretmaps line by line.

        :param secretmaps: Two secretmaps that should be compared
        :type secretmaps: ``tuple``
        :return: Void
        :rtype: ``None``
        """
        differences = difflib.unified_diff(
            self._decrypt(secretmaps[0]).splitlines(keepends=True),
            self._decrypt(secretmaps[1]).splitlines(keepends=True),
            fromfile=secretmaps[0],
            tofile=secretmaps[1]
        )

        for difference in differences:
            if difference.startswith('-'):
                click.secho(difference, nl=False, fg='red')
            elif difference.startswith('+'):
                click.secho(difference, nl=False, fg='green')
            else:
                click.secho(difference, nl=False)

    def destroy(self, treasure):
        """
        Delete secretmap file and remove key from the config file.

        :param treasure: Name of the secretmap variable
        :type treasure: ``str``
        :return: Void
        :rtype: ``None``
        """
        config = self._read_config()
        src = self._get_treasure_path(treasure)

        del config[treasure]
        self._write_config(config)
        os.remove(src)

    def render_all_files(self):
        """
        Decrypt and copy all files at the given destination.
        """
        os.makedirs(self.tmp)
        for key, value in self._read_config().items():
            basepath = os.path.dirname(value)

            with suppress(FileExistsError):
                os.makedirs(os.path.join(self.tmp, basepath))

            destination = os.path.join(self.tmp, value)
            if not os.path.exists(destination):
                self.decrypt(destination, key)

    def clear_all_files(self):
        """
        Delete all decrypted files.

        :return: Void
        :rtype: ``None``
        """
        logger.info(f'Deleting rendered secretmaps at {self.tmp}')
        shutil.rmtree(self.tmp)
        logger.info('Done!')

    def edit(self, treasure):
        """
        Edit an encrypted file.

        :param treasure: Name of the variable
        :type treasure: ``str``
        :return: Void
        :rtype: ``None``
        """
        config = self._read_config()
        filename = os.path.basename(config[treasure])
        destination = os.path.join(tempfile.gettempdir(), filename)
        self.decrypt(destination, treasure)

        editor = os.getenv('EDITOR', 'vi')
        cmd = '{} {}'.format(editor, destination)
        subprocess.run(cmd, shell=True)

        with open(destination, 'r') as f:
            tmp_file_content = f.read()

        src = self._get_treasure_path(treasure)
        with open(src, 'r') as f:
            cipher = AES256Cipher(self.keychain)
            original_file_content = cipher.decrypt(f.read())

        if original_file_content != tmp_file_content:
            self.upsert(destination, config[treasure], treasure)

        tmp_file = DeleteFile(destination)
        tmp_file.delete()

    def _get_treasure_path(self, treasure):
        """
        Get path to treasure from config file.

        :param treasure: Name of the treasure
        :type treasure: ``str``
        :return: Path to encrypted secretmap file
        :rtype: ``str``
        """
        config = self._read_config()
        try:
            return os.path.abspath(
                os.path.join(self.directory, config[treasure])
            )
        except KeyError:
            click.secho(
                f'Treasure "{treasure}" not found in {self.filepath}',
                fg='red'
            )
            sys.exit(1)

    def _read_config(self):
        """
        Load config file from the filesystem.

        :return: Content of the config file
        :rtype: ``dict``
        """
        try:
            with open(self.filepath, 'r') as f:
                config = yaml.load(f, Loader=yaml.FullLoader)
        except FileNotFoundError:
            click.secho(f'File {self.filepath} not found', fg='red')
            sys.exit(1)
        return config if config else {}

    def _write_config(self, config):
        """
        Write config file to filesystem as YAML file.

        :param config: Content of the config file
        :type config: ``dict``
        :return: Void
        :rtype: ``None``
        """
        with open(self.filepath, 'w') as f:
            f.write(yaml.dump(config))

    def _decrypt(self, treasure):
        """
        Decrypt treasure, and return it as a cleartext string.

        :param treasure: Name of the treasure
        :type treasure: ``str``
        :return: Cleartext string
        :rtype: ``str``
        """
        src = self._get_treasure_path(treasure)

        with open(src, 'r') as f:
            logger.debug(f'Decrypting secretmap on {src}')
            cipher = AES256Cipher(self.keychain)
            secretmap = cipher.decrypt(f.read())

        return secretmap



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import subprocess

import yaml
import click

from narrenschiff.log import NarrenschiffLogger


logger = NarrenschiffLogger()


class ConfigurationException(Exception):
    """Use this when something goes wrong with the configuration."""

    pass


class NarrenschiffConfiguration:
    """
    Load configuration file.

    This class should never be called directly, nor it should be inherited
    from. Other classes should use it by composition principle (see
    implementation of e.g. :class:`narrenschiff.config.Keychain`). This should
    make classes that depend on subset of config easier to test.
    """

    def __init__(self):
        conf = self._load_configuration_file()

        self.key = self._load_value(conf.get('key'))
        self.spice = self._load_value(conf.get('spice'))
        self.context = conf.get('context', {})

    def _load_value(self, path):
        """
        Load value from file on the given path.

        :param path: Path to the file
        :type path: ``str``
        """
        secret = ''
        try:
            with open(os.path.expanduser(path), 'r') as f:
                secret = f.readlines()[0].rstrip()
        except IndexError:
            click.secho(f'File {path} cannot be empty', fg='red')
            sys.exit(1)
        except FileNotFoundError:
            click.secho(f'File {path} not found', fg='red')
            click.secho(
                f'Please check or configure paths in {self._get_configuration_path()}',  # noqa
                fg='red'
            )
            sys.exit(1)
        return secret

    def _load_configuration_file(self):
        """
        Load ``.narrenschiff.yaml`` configuration file.
        """
        path = self._get_configuration_path()
        with open(path, 'r') as f:
            conf = yaml.load(f, Loader=yaml.FullLoader)
        return conf

    def _get_configuration_path(self):
        """
        Get path to the existing configuration file.

        :return: Path to configuration file
        :rtype: ``str``
        """
        project_root = os.getcwd()
        narrenschiff = os.path.join(project_root, '.narrenschiff')

        narrenschiff_yaml = '.'.join([narrenschiff, 'yaml'])
        narrenschiff_yml = '.'.join([narrenschiff, 'yml'])

        conf_yaml = self._check_configuration_file_exists(narrenschiff_yaml)
        conf_yml = self._check_configuration_file_exists(narrenschiff_yml)

        if bool(conf_yaml) != bool(conf_yml):
            return conf_yaml or conf_yml

        exception = 'Ambiguous configuration. Either missing or duplicated'
        raise ConfigurationException(exception)

    def _check_configuration_file_exists(self, path):
        """
        Check if configuration file exists. If it exists, return its path.

        :param path: Path to configuration file
        :type path: ``str``
        :return: Path
        :rtype: ``str``
        """
        if os.path.isfile(path):
            return path
        return ''


class Keychain:
    """Bundle password and salt."""

    def __init__(self):
        configuration = NarrenschiffConfiguration()

        self.key = configuration.key
        self.spice = configuration.spice


class KubectlContext:
    """Handle context switching."""

    NAME = 'undefined'
    USE = 'false'

    def __init__(self):
        configuration = NarrenschiffConfiguration()

        self.name = configuration.context.get('name', KubectlContext.NAME)
        self.use = self._sanitize_boolean(
            configuration.context.get('use', KubectlContext.USE)
        )

        if self.use:
            self.old = self._get_current_context()
            self.switch_context = (self.old, self.name)

    def _sanitize_boolean(self, value):
        if isinstance(value, str):
            if value == 'true':
                return True
        if isinstance(value, bool):
            return value
        if isinstance(value, int):
            return bool(value)
        return False

    def _get_current_context(self):
        logger.info('Obtaining current kubectl context')

        process = subprocess.run(
            'kubectl config current-context',
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        if process.returncode:
            logger.error('Could not get current context')
            logger.debug(f"{process.stderr.decode('utf-8')}")
            sys.exit(1)

        context = process.stdout.decode('utf-8').strip()
        logger.info(f'Current kubectl context is {context}')
        return context

    def switch(self):
        """Switch kubectl context."""
        old, new = self.switch_context
        logger.info(f'Switching kubectl context to {new}')

        process = subprocess.run(
            f'kubectl config use-context {new}',
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        if process.returncode:
            logger.error('Could not switch context')
            logger.debug(f"{process.stderr.decode('utf-8')}")
            sys.exit(1)

        logger.info(f'Current kubectl context is set to {new}')
        self.switch_context = (new, old)



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import click

from narrenschiff.chest import Chest
from narrenschiff.chest import AES256Cipher
from narrenschiff.config import Keychain

from narrenschiff.common import get_chest_file_path
from narrenschiff.templating import ChestVars
from narrenschiff.secretmap import CourseLocationError


@click.group()
@click.pass_context
def chest(ctx):
    """
    Load keys and spices.
    \f

    :return: Void
    :rtype: ``None``
    """
    ctx.obj = Keychain()


@chest.command()
@click.option('--treasure', help='Variable name.')
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def loot(keychain, treasure, location):
    """
    Display value from the chest file.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param location: Path to the course directory
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    path = get_chest_file_path(location)
    chest = Chest(keychain, path)
    click.echo(chest.show(treasure))


@chest.command()
@click.option('--treasure', help='Variable name.')
@click.option('--value', help='Value.')
@click.option('--location', help='Relative path to course project directory.')
@click.pass_obj
def stash(keychain, treasure, value, location):
    """
    Dynamically update chest file. Override old value if exists.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param treasure: Name of the variable
    :type treasure: ``str``
    :param value: Value of the variable
    :type value: ``str``
    :param location: Path to the course directory
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    path = get_chest_file_path(location)
    chest = Chest(keychain, path)
    chest.update(treasure, value)


@chest.command()
@click.option('--value', help='String to be encrypted.')
@click.pass_obj
def lock(keychain, value):
    """
    Encrypt string and print it to STDOUT.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param value: Value of the variable
    :type value: ``str``
    :return: Void
    :rtype: ``None``
    """
    cipher = AES256Cipher(keychain)
    click.echo(cipher.encrypt(value))


@chest.command()
@click.option('--value', help='String to be decrypted.')
@click.pass_obj
def unlock(keychain, value):
    """
    Decrypt string and print it to STDOUT.
    \f

    :param keychain: Object containing key and spice
    :type keychain: :class:`narrenschiff.chest.Keychain`
    :param value: Value of the variable
    :type value: ``str``
    :return: Void
    :rtype: ``None``
    """
    cipher = AES256Cipher(keychain)
    click.echo(cipher.decrypt(value))


@chest.command()
@click.option('--location', help='Relative path to course project directory.')
def dump(location):
    """
    Print all values from the chest on STDOUT.
    \f

    :param location: Path to the course directory
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    if not os.path.isdir(location):
        raise CourseLocationError

    vars = ChestVars(location).load_vars()

    for key, value in vars[0].items():
        print(f'{key}: {value}')



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from contextlib import suppress
from urllib.parse import urlparse

from narrenschiff.common import flatten
from narrenschiff.templating import Template
from narrenschiff.modules.mixins import KubectlDryRunMixin
from narrenschiff.modules.common import NarrenschiffModule


class Kubectl(KubectlDryRunMixin, NarrenschiffModule):
    """``kubectl`` module."""

    @property
    def cmd(self):
        command = self.command.get('command')
        switches = self.command.get('opts', [])

        self.sanitize_filenames()
        self.update_filename_argument()
        args = self.command.get('args', {})
        flags = []
        for key, value in args.items():
            flags.append("--{}='{}'".format(key, value))

        flags.extend(['--{}'.format(switch) for switch in switches])

        return ' '.join(['kubectl', command, *flags])

    def update_filename_argument(self):
        """
        Update filename argument as a properly formated string.

        :return: Void
        :rtype: ``None``
        """
        with suppress(KeyError):
            filenames = self.command['args']['filename']
            self.command['args']['filename'] = ','.join(filenames)

    def sanitize_filenames(self):
        """
        Change relative paths to absolute paths corresponding to rendered
        files.

        :return: Void
        :rtype: ``None``
        """
        with suppress(KeyError):
            filenames = flatten(list((self.command['args']['filename'],)))

            paths = []
            for filename in filenames:
                if urlparse(filename).scheme:
                    paths.append(filename)
                    continue
                paths.append(os.path.join(Template().tmp, filename))

            self.command['args']['filename'] = paths



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from jinja2 import Environment
from jinja2.exceptions import TemplateSyntaxError

import os

from narrenschiff.common import is_yaml
from narrenschiff.common import is_jinja


def _get_all_files(path):
    """
    Get all YAML files.

    :param path: Path to course project directory
    :type path: ``str``
    :return: List of valid narrenschiff files
    :rtype: ``list`` of ``str``

    Valid files can have following extensions: ``.yaml``, ``.yml``, ``.j2``,
    or ``.jinja2``.
    """
    paths = []
    for root, subdir, files in os.walk(path):
        for file in files:
            if is_yaml(file) or is_jinja(file):
                paths.append(os.path.join(root, file))
    return paths


def _validate_template(path):
    """
    Validate jinja template.

    :param path: Path to course project directory
    :type path: ``str``
    :return: List of valid narrenschiff files
    :rtype: ``list`` of ``str``
    """
    env = Environment()

    try:
        with open(path) as f:
            env.parse(f.read())
    except TemplateSyntaxError as e:
        click.echo(f'\033[35m{path}\033[0m: {e}')
        return 1
    return 0


@click.command()
@click.option('--location', help='Path to the directory.')
@click.pass_context
def lint(ctx, location):
    """
    Lint project files. Check if they are valid Jinja2 templates.
    \f

    :param location: Path to the directory
    :type location: ``str``
    :return: Void
    :rtype: ``None``
    """
    results = []
    for file in _get_all_files(os.path.abspath(location)):
        results.append(_validate_template(file))
    ctx.exit(code=int(any(results)))



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

import os
import re
import platform
import subprocess
from shutil import which

import narrenschiff


def _check_cmd(cmd, subcommand, formatted):
    """Check if command exists on the PATH."""
    if not which(f'{cmd}'):
        click.echo(f'* {cmd}: Not found')
    else:
        click.echo(f'* {cmd}')
        if formatted:
            click.echo('```')
        subprocess.run(f'{cmd} {subcommand}', shell=True)
        if formatted:
            click.echo('```')


def _dependency_management(formatted):
    """Detect all dependency management tools."""
    tools = ['pip', 'pipenv', 'pip-compile', 'pip-sync', 'poetry']
    detected = []
    for tool in tools:
        check_tool = subprocess.run(f'{tool} --version',
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        if check_tool.returncode == 0:
            version = re.search(
                r'\d+\.\d+\.\d+',
                check_tool.stdout.decode().strip()
            ).group()
        else:
            version = 'Not Found'
        detected.append({
            'tool': tool,
            'version': version
        })

    click.echo('* Dependency management')
    if formatted:
        click.echo('\n| Dependency management tool | Version |')
        click.echo('| :------------------------: | :-----: |')

    for tool in detected:
        if formatted:
            click.echo(f'| {tool["tool"]} | {tool["version"]} |')
        else:
            click.echo(f'{tool["tool"]}: {tool["version"]}')


@click.command()
@click.option('--formatted',
              required=False,
              is_flag=True,
              help='Output in Markdown format.')
def env(formatted):
    """Show environment info."""
    click.echo(f'* narrenschiff: {narrenschiff.__version__}')
    click.echo(f'* OS: {platform.platform()}')
    _check_cmd('python', '-VV', formatted)
    _dependency_management(formatted)
    click.echo(f'* virtualenv: {"yes" if os.getenv("VIRTUAL_ENV") else "no"}')
    _check_cmd('gcloud', 'version', formatted)
    _check_cmd('kubectl', 'version', formatted)
    _check_cmd('helm', 'version', formatted)



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import datetime
import click


class AmbiguousOptions(Exception):
    """Use when tasks have undefined YAML tags."""

    pass


class Task:
    """
    Parse the smallest unit of the course.

    Each dictionary item from the task list should be wrapped with this class::

        course = [{'name': 'Kustomize', 'kustomization': 'examples/app/'}]
        tasks = [Task(t) for t in course]

        task = tasks[0]
        task.name  # --> 'Kustomize'
        taks.kustomization  # --> instance of Kustomization class
    """

    def __init__(self, task):
        self.name = task.pop('name', None)
        self.beacons = set(task.pop('beacons', []))

        if len(task) > 1:
            options = ', '.join(task.keys())
            raise AmbiguousOptions('Check tags for unknown options:', options)

        # The remaining YAML tag should be command
        command_name = list(task.keys())[0]
        command_args = task.pop(command_name)
        Command = self._dynamic_module_import(command_name)
        self.command = Command(command_args)

    def __str__(self):
        return self.name

    def __repr__(self):
        module = self.__class__.__module__
        name = self.__class__.__name__
        return '<{}.{} ({})>'.format(module, name, self.name)

    def _dynamic_module_import(self, module):
        """
        Import class corresponding to YAML tag.

        :param module: Name of the command
        :type module: ``str``
        :return: Class representing the command
        :rtype: ``class``

        **Important:** A *module* in this context means the "`narrenschiff`"
        module. This is a module that deals with executing the command
        described in ``tasks.yaml`` file i.e. the "`course`" file.
        """
        path = 'narrenschiff.modules.{}'.format(module.lower())
        klass = ''.join(
            '{}{}'.format(s[0].capitalize(), s[1:]) for s in module.split('_')
        )
        mod = __import__(path, fromlist=[klass])
        return getattr(mod, klass)


class TasksEngine:
    """Run course."""

    def __init__(self, tasks, beacons, dry_run_enabled):
        """
        Construct a :class:`narrenschiff.task.TasksEngine` class.

        :param tasks: List of tasks
        :type tasks: ``list`` of :class:`narrenschiff.task.Task`
        :param beacons: List of tags used to determine which task is run
        :type beacons: ``set``
        :param dry_run_enabled: Boolean indicating whether user turned on dry
            run for the task
        :type dry_run_enabled: ``bool``
        :return: Void
        :rtype: ``None``
        """
        self.tasks = tasks
        self.beacons = beacons
        self.dry_run_enabled = dry_run_enabled
        self.width = int(subprocess.check_output(['tput', 'cols']).decode())

    def run(self):
        """
        Start executing tasks.

        :return: Void
        :rtype: ``None``
        """
        click.echo()
        for task in self.tasks:
            if self.beacons:
                if self.beacons & task.beacons or 'always' in task.beacons:
                    self._execute(task)
            else:
                self._execute(task)

    def _execute(self, task):
        """
        Execute a task.

        :param task: Task that is about to get executed
        :type task: :class:`narrenschiff.task.Task`
        :return: Void
        :rtype: ``None``
        """
        width = int(self.width) - 41 - len(task.name)
        current_time = datetime.datetime.now()
        fill = '*' * width
        click.echo(
            '* [ {} ] * [ {} ] {}\n'.format(current_time, task.name, fill)
        )
        task.command.execute(dry_run_enabled=self.dry_run_enabled)



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from narrenschiff.modules.common import NarrenschiffModule


class Gcloud(NarrenschiffModule):
    """``gcloud`` module."""

    @property
    def cmd(self):
        command = self.command.get('command')
        args = self.command.get('args', {})
        switches = self.command.get('opts', [])

        flags = []
        for flag, value in args.items():
            flags.append("--{} '{}'".format(flag, args[flag]))

        flags.extend(['--{}'.format(switch) for switch in switches])

        return ' '.join(['gcloud', command, *flags])

    def dry_run_supported(self, cmd):
        return False



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config
from colorlog import ColoredFormatter

from narrenschiff.common import Singleton


LOGGING = {
    'version': 1,
    'handlers': {
        'null': {
            'class': 'logging.NullHandler',
        },
    },
    'loggers': {
        'narrenschiff': {
            'handlers': ['null'],
            'propagate': True,
            'level': 'INFO',
        },
    }
}

logging.config.dictConfig(LOGGING)


class NarrenschiffLogger(metaclass=Singleton):
    """Set narrenschiff log level and print to STDOUT."""

    LOG_LEVEL = {
        5: logging.DEBUG,
        4: logging.INFO,
        3: logging.WARNING,
        2: logging.ERROR,
        1: logging.CRITICAL
    }

    def __init__(self):
        self.logger = logging.getLogger('narrenschiff')

    def set_verbosity(self, verbosity):
        """
        Set verbosity of the logger.

        :param verbosity: Verbosity level
        :type verbosity: ``int``

        Verbosity can be from 1 to 5, corresponding to five log levels.
        """
        if verbosity not in range(1, 6):
            return
        formatter = ColoredFormatter(
            '%(log_color)s%(levelname)s %(asctime)s %(message)s',
            log_colors={
                'DEBUG': 'blue',
                'INFO': 'white',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red,bold',
            },
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(NarrenschiffLogger.LOG_LEVEL.get(verbosity))

    def __getattr__(self, name):
        log_levels = ['debug', 'info', 'warning', 'error', 'critical']
        if name in log_levels:
            return getattr(self.logger, name)



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from contextlib import suppress

from narrenschiff.modules.common import NarrenschiffModule
from narrenschiff.secretmap import Secretmap


class HelmException(Exception):
    """Raise when something is wrong with Helm module."""

    pass


class Helm(NarrenschiffModule):
    """``helm`` module."""

    @property
    def cmd(self):
        command = self.command.get('command')
        name = self.command.get('name', '')
        chart = self.command.get('chart', '')
        options = self.command.get('opts', [])
        arguments = self.command.get('args', {})

        self.parse_secretmaps_args()

        try:
            args_set = self.command.get('args').get('set')
        except AttributeError:
            args_set = None

        opts = ''
        if options:
            opts = ' '.join(map(lambda opt: '--{}'.format(opt), options))

        with suppress(AttributeError, TypeError):
            values = ','.join(self.command.get('args').get('values'))
            self.command['args']['values'] = values

        args = ' '.join(
            ['--{} {}'.format(key, value)
                for key, value in arguments.items()
                if key != 'set']
        )

        sets = ''
        if args_set:
            sets = ' '.join(['--set {}'.format(s) for s in args_set])

        return ' '.join(['helm', command, name, chart, opts, args, sets])

    def parse_secretmaps_args(self):
        """
        Mutate secretmap arguments. Expand secretmap paths to match files in
        the ``/tmp`` directory.

        :return: Void
        :rtype: ``None``
        """
        for key, value in self.command.get('args', {}).items():
            try:
                self.command['args'][key] = self._template_path(value)
            except AttributeError:
                if key == 'values':
                    values = [self._template_path(v) for v in value]
                    self.command['args'][key] = values

    def _template_path(self, value):
        """Replace secretmap filter artifact with `/tmp` file path."""
        secretmap = '{{secretmap}}/'
        if value.startswith(secretmap):
            basepath = value.replace(secretmap, '')
            tmp = Secretmap().tmp
            return os.path.join(tmp, basepath)
        return value

    def dry_run_supported(self, cmd):
        whitelist = [
            'install',
            'template',
            'uninstall',
            'upgrade',
        ]

        if cmd.split()[1] in whitelist:
            return True
        return False



# Copyright 2021 The Narrenschiff Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import subprocess
from abc import ABC
from abc import abstractmethod

import click

from narrenschiff.log import NarrenschiffLogger


logger = NarrenschiffLogger()


class NarrenschiffModuleException(Exception):
    """Use when something goes wrong in modules."""

    pass


class NarrenschiffModule(ABC):
    """
    Abstract class/Interface for the module classes. A module must inherit
    from this class.

    :cvar DRY_RUN_FLAG: ``int``

    The subprocess module does not have a way of indicating whether a command
    was run with dry run or not, since that is the responsibility of
    :meth:`narrenschiff.modules.common.NarrenschiffModule.execute` method. If
    module or subcommand of module is not supporting dry run, than
    ``DRY_RUN_FLAG`` is a reserved return code (rc) that indicates that program
    should not exit, and output should be printed in special color (blue).
    """

    DRY_RUN_FLAG = -99999

    def __init__(self, command):
        """
        Construct a module that executes command.

        :param command: Arguments for the module used to construct a command
        :type command: ``str``, ``int``, ``list``, or ``dict``
        :return: Void
        :rtype: ``None``
        """
        self.command = command

    def __str__(self):
        return self.__class__.__name__.lower()

    def __repr__(self):
        module = self.__class__.__module__
        name = self.__class__.__name__
        return '<{}.{} object at {}>'.format(module, name, hex(id(self)))

    def execute(self, dry_run_enabled=False):
        """
        Parse command and its arguments, and execute the module.

        :param dry_run_enabled: Boolean indicating whether user turned on dry
            run for the task
        :type dry_run_enabled: ``bool``
        :return: Void
        :rtype: ``None``
        """
        if dry_run_enabled:
            if self.dry_run_supported(self.cmd):
                output, rc = self.subprocess(f'{self.cmd} {self.dry_run}')
            else:
                output, rc = (
                    'Dry run not supported by the module or a subcommand\n',
                    NarrenschiffModule.DRY_RUN_FLAG
                )
        else:
            output, rc = self.subprocess(self.cmd)
        self.echo(output, rc)

    @property
    @abstractmethod
    def cmd(self):
        """
        Get command that module needs to execute later.

        :return: Full command with all parameters
        :rtype: ``str``
        """
        raise NotImplementedError

    @property
    def dry_run(self):
        """
        Return a dry run flag.

        :return: ``--dry-run`` string
        :rtype: ``str``

        In general most commands use ``--dry-run`` so there is no need to
        override this. However, there are exceptions for some commands where
        this flag is differently named. This property offers extensibility to
        the modules that may use different flag.
        """
        return '--dry-run'

    @abstractmethod
    def dry_run_supported(self, cmd):
        """
        Check if command supports --dry-run.

        :param cmd: Command that module should execute
        :type cmd: ``str``
        :return: Boolean indicating whether command supports dry run
        :rtype: ``bool``
        """
        raise NotImplementedError

    def subprocess(self, cmd):
        """
        Execute command with shell, and return output and return code.

        :param cmd: Command to execute
        :type cmd: ``str``
        :return: Output and return code
        :rtype: ``tuple``

        Example::

            output, rc = self.subprocess('kubectl get pods')
        """
        process = subprocess.run(
            cmd,
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        output = process.stderr if process.stderr else process.stdout

        logger.info(f'Command "{cmd}" executed')

        return output.decode('utf-8'), process.returncode

    def echo(self, output, rc):
        """
        Print output to console, and exit if return code is different from 0.

        :param output: stdout or stderr of a process
        :type output: ``str``
        :param rc: Return code of the process
        :type rc: ``int``
        :return: Void
        :rtype: ``None``
        """
        color = self._color(rc)

        if output == '' and rc == 0:
            # No output from the task, but operation was successful
            output = 'Operation successfully executed!'

        click.secho(output, fg=color)

        if rc and rc != NarrenschiffModule.DRY_RUN_FLAG:
            sys.exit(rc)

    def _color(self, rc):
        """
        Get color for the command output.

        :param rc: Return code of the command
        :type rc: ``int``
        :return: String indicating color
        :rtype: ``str``
        """
        if rc == 0:
            return 'green'
        elif rc == NarrenschiffModule.DRY_RUN_FLAG:
            return 'blue'
        else:
            return 'red'
